/***************************************************************************
 * This program is Copyright (C) 1986, 1987, 1988 by Jonathan Payne.  JOVE *
 * is provided to you without charge, and with no warranty.  You may give  *
 * away copies of JOVE, including sources, provided that this notice is    *
 * included in all the files.                                              *
 ***************************************************************************/

/* Commands: */

extern void
	CasRegLower proto((void)),
	CasRegUpper proto((void)),
	CapChar proto((void)),
	CapWord proto((void)),
	LowWord proto((void)),
	UppWord proto((void));
